(this.webpackJsonpstreamlit_component_template=this.webpackJsonpstreamlit_component_template||[]).push([[0],{16:function(e,t,n){e.exports=n(30)},30:function(e,t,n){"use strict";n.r(t);var r=n(6),a=n.n(r),o=n(13),c=n.n(o),l=n(0),p=n(1),s=n(2),i=n(3),u=n(9),m=n(15),v=n.n(m),d=function(e){Object(s.a)(n,e);var t=Object(i.a)(n);function n(){var e;Object(p.a)(this,n);for(var r=arguments.length,o=new Array(r),c=0;c<r;c++)o[c]=arguments[c];return(e=t.call.apply(t,[this].concat(o))).render=function(){var t=JSON.parse(e.props.args.plot_obj),n=JSON.parse(e.props.args.config),r=e.props.args.override_height,o=e.props.args.override_width,c=e.props.args.click_event,l=e.props.args.select_event,p=e.props.args.hover_event;return u.a.setFrameHeight(r),a.a.createElement(v.a,{data:t.data,layout:t.layout,config:n,frames:t.frames,onClick:c?e.plotlyEventHandler:function(){},onSelected:l?e.plotlyEventHandler:function(){},onHover:p?e.plotlyEventHandler:function(){},style:{width:o,height:r},className:"stPlotlyChart"})},e.plotlyEventHandler=function(e){var t=[];e.points.forEach((function(e){t.push({x:e.x,y:e.y,curveNumber:e.curveNumber,pointNumber:e.pointNumber,pointIndex:e.pointIndex})})),u.a.setComponentValue(JSON.stringify(t))},e}return Object(l.a)(n)}(u.b),f=Object(u.c)(d);c.a.render(a.a.createElement(a.a.StrictMode,null,a.a.createElement(f,null)),document.getElementById("root"))}},[[16,1,2]]]);
//# sourceMappingURL=main.2ba31fb4.chunk.js.map