self.__precacheManifest = (self.__precacheManifest || []).concat([
  {
    "revision": "0dc326b0b7b5714d0d951b6041652040",
    "url": "./index.html"
  },
  {
    "revision": "935abba20d5d8dc289c9",
    "url": "./static/js/2.a9de1b85.chunk.js"
  },
  {
    "revision": "4c73541be880e0770f8c0222295bcc32",
    "url": "./static/js/2.a9de1b85.chunk.js.LICENSE.txt"
  },
  {
    "revision": "5ed8974c0b5a59828672",
    "url": "./static/js/main.2ba31fb4.chunk.js"
  },
  {
    "revision": "7c26bca7e16783d14d15",
    "url": "./static/js/runtime-main.11ec9aca.js"
  }
]);